import contextlib
import os
from contextlib import contextmanager
from functools import wraps
from collections.abc import Sequence as C_Sequence
from typing import Any, Callable, ClassVar, Dict, List, Iterator, Mapping, Optional, Sequence, Tuple, get_type_hints, overload

from dlt.common import json, logger, signals, pendulum
from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ConfigFieldMissingException
from dlt.common.configuration.specs.config_namespace_context import ConfigNamespacesContext
from dlt.common.exceptions import MissingDependencyException
from dlt.common.runners.runnable import Runnable
from dlt.common.schema.exceptions import InvalidDatasetName
from dlt.common.schema.typing import TColumnSchema, TWriteDisposition
from dlt.common.schema.utils import default_normalizers, import_normalizers
from dlt.common.storages.load_storage import LoadStorage
from dlt.common.typing import TFun, TSecretValue

from dlt.common.runners import pool_runner as runner, TRunMetrics, initialize_runner
from dlt.common.storages import LiveSchemaStorage, NormalizeStorage

from dlt.common.configuration import inject_namespace
from dlt.common.configuration.specs import RunConfiguration, NormalizeVolumeConfiguration, SchemaVolumeConfiguration, LoadVolumeConfiguration, PoolRunnerConfiguration
from dlt.common.destination import DestinationCapabilitiesContext, DestinationReference, JobClientBase, DestinationClientConfiguration, DestinationClientDwhConfiguration, TDestinationReferenceArg
from dlt.common.pipeline import LoadInfo, TPipelineState
from dlt.common.schema import Schema
from dlt.common.storages.file_storage import FileStorage
from dlt.common.utils import is_interactive

from dlt.extract.exceptions import SourceExhausted
from dlt.extract.extract import ExtractorStorage, extract
from dlt.extract.source import DltResource, DltSource
from dlt.normalize import Normalize
from dlt.normalize.configuration import NormalizeConfiguration
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.job_client_impl import SqlJobClientBase
from dlt.load.configuration import LoaderConfiguration
from dlt.load import Load

from dlt.pipeline.exceptions import CannotRestorePipelineException, InvalidPipelineName, PipelineConfigMissing, PipelineStepFailed, SqlClientNotAvailable
from dlt.pipeline.typing import TPipelineStep
from dlt.pipeline.state import STATE_ENGINE_VERSION, load_state_from_destination, merge_state_if_changed, state_resource, StateInjectableContext


def with_state_sync(extract_state: bool = False) -> Callable[[TFun], TFun]:

    def decorator(f: TFun) -> TFun:
        @wraps(f)
        def _wrap(self: "Pipeline", *args: Any, **kwargs: Any) -> Any:
            # backup and restore state
            with self._managed_state(extract_state=extract_state) as state:
                # add the state to container as a context
                with self._container.injectable_context(StateInjectableContext(state=state)):
                    return f(self, *args, **kwargs)

        return _wrap  # type: ignore

    return decorator

def with_schemas_sync(f: TFun) -> TFun:

    @wraps(f)
    def _wrap(self: "Pipeline", *args: Any, **kwargs: Any) -> Any:
        for name in self._schema_storage.live_schemas:
            # refresh live schemas in storage or import schema path
            self._schema_storage.commit_live_schema(name)
        return f(self, *args, **kwargs)

    return _wrap  # type: ignore

def with_config_namespace(namespaces: Tuple[str, ...]) -> Callable[[TFun], TFun]:

    def decorator(f: TFun) -> TFun:

        @wraps(f)
        def _wrap(self: "Pipeline", *args: Any, **kwargs: Any) -> Any:
            # add namespace context to the container to be used by all configuration without explicit namespaces resolution
            with inject_namespace(ConfigNamespacesContext(pipeline_name=self.pipeline_name, namespaces=namespaces)):
                return f(self, *args, **kwargs)

        return _wrap  # type: ignore

    return decorator


class Pipeline:

    STATE_FILE: ClassVar[str] = "state.json"
    STATE_PROPS: ClassVar[List[str]] = list(get_type_hints(TPipelineState).keys())

    pipeline_name: str
    default_schema_name: str = None
    schema_names: List[str] = []
    full_refresh: bool
    working_dir: str
    pipeline_root: str
    destination: DestinationReference = None
    dataset_name: str = None
    credentials: Any = None

    def __init__(
            self,
            pipeline_name: str,
            working_dir: str,
            pipeline_salt: TSecretValue,
            destination: DestinationReference,
            dataset_name: str,
            credentials: Any,
            import_schema_path: str,
            export_schema_path: str,
            full_refresh: bool,
            must_attach_to_local_pipeline: bool,
            restore_from_destination: bool,
            runtime: RunConfiguration
        ) -> None:
        self.pipeline_salt = pipeline_salt
        self.runtime_config = runtime
        self.full_refresh = full_refresh

        self._container = Container()
        # TODO: pass default normalizers as context or as config with defaults
        self._default_naming, _ = import_normalizers(default_normalizers())
        self._pipeline_instance_id = pendulum.now().format("_YYYYMMDDhhmmss")
        self._pipeline_storage: FileStorage = None
        self._schema_storage: LiveSchemaStorage = None
        self._schema_storage_config: SchemaVolumeConfiguration = None
        self._normalize_storage_config: NormalizeVolumeConfiguration = None
        self._load_storage_config: LoadVolumeConfiguration = None

        initialize_runner(self.runtime_config)
        # initialize pipeline working dir
        self._init_working_dir(pipeline_name, working_dir)

        # set destination and dataset, those may be overwritten by state but we'll need them to eventually restore state
        self.destination = destination or self.destination  # changing the destination could be dangerous if pipeline has not loaded items
        self._set_dataset_name(dataset_name)
        self.credentials = credentials

        # do not restore from destination if full refresh or must must attach to local working folder
        restore_from_destination = not must_attach_to_local_pipeline and not full_refresh and restore_from_destination
        # require dataset name to be present if restoring
        if restore_from_destination and not dataset_name:
            raise PipelineConfigMissing(self.pipeline_name, "dataset_name", "init", "When requesting the pipeline to be restored from the destination, 'dataset_name` must be passed explicitly or configured via config providers.")
        with self._managed_state(restore_from_destination=restore_from_destination) as state:
            # see if state didn't change the pipeline name
            if "pipeline_name" in state and pipeline_name != state["pipeline_name"]:
                raise CannotRestorePipelineException(pipeline_name, working_dir, f"working directory contains state for pipeline with name {self.pipeline_name}")
            # at this moment state is recovered so we overwrite the state with the values from init
            self.destination = destination or self.destination  # changing the destination could be dangerous if pipeline has not loaded items
            self._set_dataset_name(dataset_name)
            self._configure(import_schema_path, export_schema_path, must_attach_to_local_pipeline)
            self._restore_schemas_from_destination()

    def drop(self) -> "Pipeline":
        """Deletes existing pipeline state, schemas and drops datasets at the destination if present"""
        # if self.destination:
        #     # drop the data for all known schemas
        #     for schema in self._schema_storage:
        #         with self._get_destination_client(self._schema_storage.load_schema(schema)) as client:
        #             client.initialize_storage(wipe_data=True)
        # reset the pipeline working dir
        self._create_pipeline()
        # clone the pipeline
        return Pipeline(
            self.pipeline_name,
            self.working_dir,
            self.pipeline_salt,
            self.destination,
            self.dataset_name,
            self.credentials,
            self._schema_storage.config.import_schema_path,
            self._schema_storage.config.export_schema_path,
            self.full_refresh,
            False,
            False,
            self.runtime_config
        )


    # @overload
    # def extract(
    #     self,
    #     data: Union[Iterator[TResolvableDataItem], Iterable[TResolvableDataItem]],
    #     table_name = None,
    #     write_disposition = None,
    #     parent = None,
    #     columns = None,
    #     max_parallel_data_items: int =  20,
    #     schema: Schema = None
    # ) -> None:
    #     ...

    # @overload
    # def extract(
    #     self,
    #     data: DltSource,
    #     max_parallel_iterators: int = 1,
    #     max_parallel_data_items: int =  20,
    #     schema: Schema = None
    # ) -> None:
    #     ...

    @with_schemas_sync  # this must precede with_state_sync
    @with_state_sync(extract_state=True)  # extract pipeline state with the package
    @with_config_namespace(("extract",))
    def extract(
        self,
        data: Any,
        *,
        table_name: str = None,
        parent_table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None,
        max_parallel_items: int = 100,
        workers: int = 5
    ) -> None:

        def apply_hint_args(resource: DltResource) -> None:
            columns_dict = None
            if columns:
                columns_dict = {c["name"]:c for c in columns}
            # apply hints only if any of the hints is present, table_name must be always present
            if table_name or parent_table_name or write_disposition or columns:
                resource.apply_hints(table_name or resource.name, parent_table_name, write_disposition, columns_dict)

        def choose_schema() -> Schema:
            if schema:
                return schema
            if self.default_schema_name:
                return self.default_schema
            return self._make_schema_with_default_name()

        effective_schema = choose_schema()

        # a list of sources or a list of resources may be passed as data
        sources: List[DltSource] = []

        def item_to_source(data_item: Any) -> DltSource:
            if isinstance(data_item, DltSource):
                # if schema is explicit then override source schema
                if schema:
                    data_item.schema = schema
                # try to apply hints to resources
                resources = data_item.resources.values()
                for r in resources:
                    apply_hint_args(r)
                return data_item

            if isinstance(data_item, DltResource):
                # apply hints
                apply_hint_args(data_item)
                # package resource in source
                return DltSource(effective_schema.name, effective_schema, [data_item])

            # iterator/iterable/generator
            # create resource first without table template
            resource = DltResource.from_data(data_item, name=table_name)
            # apply hints
            apply_hint_args(resource)
            # wrap resource in source
            return DltSource(effective_schema.name, effective_schema, [resource])

        if isinstance(data, C_Sequence) and len(data) > 0:
            # if first element is source or resource
            if isinstance(data[0], DltResource):
                sources.append(item_to_source(DltSource(effective_schema.name, effective_schema, data)))
            elif isinstance(data[0], DltSource):
                for s in data:
                    sources.append(item_to_source(s))
            else:
                sources.append(item_to_source(data))
        else:
            sources.append(item_to_source(data))

        try:
            # extract all sources
            for s in sources:
                if s.exhausted:
                    raise SourceExhausted(s.name)
                self._extract_source(s, max_parallel_items, workers)
        except Exception as exc:
            # TODO: provide metrics from extractor
            raise PipelineStepFailed(self.pipeline_name, "extract", exc, runner.LAST_RUN_METRICS) from exc


    @with_schemas_sync
    @with_config_namespace(("normalize",))
    def normalize(self, workers: int = 1) -> None:
        if is_interactive() and workers > 1:
            raise NotImplementedError("Do not use normalize workers in interactive mode ie. in notebook")
        # check if any schema is present, if not then no data was extracted
        if not self.default_schema_name:
            return

        # get destination capabilities
        destination_caps = self._get_destination_capabilities()
        # create default normalize config
        normalize_config = NormalizeConfiguration(
            is_single_run=True,
            exit_on_exception=True,
            workers=workers,
            pool_type="none" if workers == 1 else "process",
            _schema_storage_config=self._schema_storage_config,
            _normalize_storage_config=self._normalize_storage_config,
            _load_storage_config=self._load_storage_config
        )
        # run with destination context
        with self._container.injectable_context(destination_caps):
            # shares schema storage with the pipeline so we do not need to install
            normalize = Normalize(config=normalize_config, schema_storage=self._schema_storage)
            self._run_step_in_pool("normalize", normalize, normalize.config)

    @with_state_sync()
    @with_schemas_sync
    @with_config_namespace(("load",))
    def load(
        self,
        destination: DestinationReference = None,
        dataset_name: str = None,
        credentials: Any = None,
        # raise_on_failed_jobs = False,
        # raise_on_incompatible_schema = False,
        # always_wipe_storage: bool = False,
        *,
        workers: int = 20
    ) -> LoadInfo:

        # if destination is a string then resolve it
        if isinstance(destination, str):
            destination = DestinationReference.from_name(destination)

        # set destination and default dataset if provided
        self.destination = destination or self.destination
        self._set_dataset_name(dataset_name)
        self.credentials = credentials
        # check if any schema is present, if not then no data was extracted
        if not self.default_schema_name:
            return None

        # make sure that destination is set and client is importable and can be instantiated
        client_initial_config = self._get_destination_client_initial_config()
        self._get_destination_client(self.default_schema, client_initial_config)

        # create initial loader config and the loader
        load_config = LoaderConfiguration(
            is_single_run=True,
            exit_on_exception=True,
            workers=workers,
            always_wipe_storage=False,
            _load_storage_config=self._load_storage_config
        )
        load = Load(self.destination, is_storage_owner=False, config=load_config, initial_client_config=client_initial_config)
        try:
            self._run_step_in_pool("load", load, load.config)
            return self._get_load_info(load)
        except PipelineStepFailed as pipex:
            pipex.step_info = self._get_load_info(load)
            raise

    @with_config_namespace(("run",))
    def run(
        self,
        data: Any = None,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None
    ) -> LoadInfo:
        # set destination and default dataset if provided
        destination = DestinationReference.from_name(destination)
        self.destination = destination or self.destination
        self._set_dataset_name(dataset_name)

        # normalize and load pending data
        self.normalize()
        info = self.load(destination, dataset_name, credentials=credentials)

        # extract from the source
        if data:
            self.extract(data, table_name=table_name, write_disposition=write_disposition, columns=columns, schema=schema)
            self.normalize()
            return self.load(destination, dataset_name, credentials=credentials)
        else:
            return info

    @property
    def schemas(self) -> Mapping[str, Schema]:
        return self._schema_storage

    @property
    def default_schema(self) -> Schema:
        return self.schemas[self.default_schema_name]

    @property
    def state(self) -> TPipelineState:
        return self._get_state()

    @property
    def last_run_exception(self) -> BaseException:
        return runner.LAST_RUN_EXCEPTION

    def list_extracted_resources(self) -> Sequence[str]:
        return self._get_normalize_storage().list_files_to_normalize_sorted()

    def list_normalized_load_packages(self) -> Sequence[str]:
        return self._get_load_storage().list_packages()

    def list_completed_load_packages(self) -> Sequence[str]:
        return self._get_load_storage().list_completed_packages()

    def list_failed_jobs_in_package(self, load_id: str) -> Sequence[Tuple[str, str]]:
        storage = self._get_load_storage()
        failed_jobs: List[Tuple[str, str]] = []
        for file in storage.list_completed_failed_jobs(load_id):
            if not file.endswith(".exception"):
                try:
                    failed_message = storage.storage.load(file + ".exception")
                except FileNotFoundError:
                    failed_message = None
                failed_jobs.append((storage.storage.make_full_path(file), failed_message))
        return failed_jobs

    def sync_schema(self, schema_name: str = None, credentials: Any = None) -> None:
        schema = self.schemas[schema_name] if schema_name else self.default_schema
        with self._get_destination_client(schema, self._get_destination_client_initial_config(credentials)) as client:
            client.initialize_storage()
            client.update_storage_schema()

    def sql_client(self, schema_name: str = None, credentials: Any = None) -> SqlClientBase[Any]:
        if not self.default_schema_name:
            raise PipelineConfigMissing(self.pipeline_name, "default_schema_name", "load", "Sql Client is not available in a pipeline without a default schema. Extract some data first or restore the pipeline from the destination using 'restore_from_destination' flag. There's also `_inject_schema` method for advanced users.")
        schema = self.schemas[schema_name] if schema_name else self.default_schema
        return self._sql_job_client(schema, credentials).sql_client

    def _sql_job_client(self, schema: Schema, credentials: Any = None) -> SqlJobClientBase:
        client = self._get_destination_client(schema, self._get_destination_client_initial_config(credentials))
        if isinstance(client, SqlJobClientBase):
            return client
        else:
            raise SqlClientNotAvailable(self.pipeline_name, self.destination.__name__)

    def _get_normalize_storage(self) -> NormalizeStorage:
        return NormalizeStorage(True, self._normalize_storage_config)

    def _get_load_storage(self) -> LoadStorage:
        caps = self._get_destination_capabilities()
        return LoadStorage(True, caps.preferred_loader_file_format, caps.supported_loader_file_formats, self._load_storage_config)

    def _init_working_dir(self, pipeline_name: str, working_dir: str) -> None:
        self.pipeline_name = pipeline_name
        self.working_dir = working_dir
        self._validate_pipeline_name()
        # compute the folder that keeps all of the pipeline state
        self.pipeline_root = os.path.join(working_dir, pipeline_name)
        # create pipeline storage, do not create working dir yet
        self._pipeline_storage = FileStorage(self.pipeline_root, makedirs=False)
        # if full refresh was requested, wipe out all data from working folder, if exists
        if self.full_refresh:
            self._wipe_working_folder()

    def _configure(self, import_schema_path: str, export_schema_path: str, must_attach_to_local_pipeline: bool) -> None:
        # create schema storage and folders
        self._schema_storage_config = SchemaVolumeConfiguration(
            schema_volume_path=os.path.join(self.pipeline_root, "schemas"),
            import_schema_path=import_schema_path,
            export_schema_path=export_schema_path
        )
        # create default configs
        self._normalize_storage_config = NormalizeVolumeConfiguration(normalize_volume_path=os.path.join(self.pipeline_root, "normalize"))
        self._load_storage_config = LoadVolumeConfiguration(load_volume_path=os.path.join(self.pipeline_root, "load"),)

        # are we running again?
        has_state = self._pipeline_storage.has_file(Pipeline.STATE_FILE)
        if must_attach_to_local_pipeline and not has_state:
            raise CannotRestorePipelineException(self.pipeline_name, self.working_dir, f"the pipeline was not found in {self.pipeline_root}.")

        # attach to pipeline if folder exists and contains state
        if has_state:
            self._attach_pipeline()
        else:
            # this will erase the existing working folder
            self._create_pipeline()

        # create schema storage
        self._schema_storage = LiveSchemaStorage(self._schema_storage_config, makedirs=True)

    def _create_pipeline(self) -> None:
        self._wipe_working_folder()
        self._pipeline_storage.create_folder("", exists_ok=False)

    def _wipe_working_folder(self) -> None:
        # kill everything inside the working folder
        if self._pipeline_storage.has_folder(""):
            self._pipeline_storage.delete_folder("", recursively=True)

    def _attach_pipeline(self) -> None:
        pass

    def _extract_source(self, source: DltSource, max_parallel_items: int, workers: int) -> None:
        # discover the schema from source
        source_schema = source.discover_schema()
        pipeline_schema: Schema = None
        with contextlib.suppress(FileNotFoundError):
            pipeline_schema = self._schema_storage.load_schema(source_schema.name)
        should_initialize_import = False

        # if source schema does not exist in the pipeline
        if source_schema.name not in self._schema_storage:
            # possibly initialize the import schema if it is a new schema
            should_initialize_import = True
            # save schema into the pipeline
            self._schema_storage.save_schema(source_schema)
        pipeline_schema = self._schema_storage[source_schema.name]

        # and set as default if this is first schema in pipeline
        if not self.default_schema_name:
            # this performs additional validations as schema contains the naming module
            self._set_default_schema_name(source_schema)

        # get the current schema and merge tables from source_schema, we'll not merge the high level props
        for table in source_schema.all_tables():
            pipeline_schema.update_schema(pipeline_schema.normalize_table_identifiers(table))

        self._iterate_source(source, pipeline_schema, max_parallel_items, workers)

        # initialize import with fully discovered schema
        if should_initialize_import:
            self._schema_storage.initialize_import_schema(pipeline_schema)

    def _iterate_source(self, source: DltSource, pipeline_schema: Schema, max_parallel_items: int, workers: int) -> None:
        # TODO: create source context
        # iterate over all items in the pipeline and update the schema if dynamic table hints were present
        storage = ExtractorStorage(self._normalize_storage_config)
        # inject the config namespace with the current source name
        with inject_namespace(ConfigNamespacesContext(namespaces=("sources", source.name))):
            extractor = extract(source, storage, max_parallel_items=max_parallel_items, workers=workers)
            # source iterates
            source.exhausted = True
            for _, partials in extractor.items():
                for partial in partials:
                    pipeline_schema.update_schema(pipeline_schema.normalize_table_identifiers(partial))

    def _run_step_in_pool(self, step: TPipelineStep, runnable: Runnable[Any], config: PoolRunnerConfiguration) -> int:
        try:
            ec = runner.run_pool(config, runnable)
            # in any other case we raise if runner exited with status failed
            if runner.LAST_RUN_METRICS.has_failed:
                raise PipelineStepFailed(self.pipeline_name, step, self.last_run_exception, runner.LAST_RUN_METRICS)
            return ec
        except Exception as r_ex:
            # if EXIT_ON_EXCEPTION flag is set, exception will bubble up directly
            raise PipelineStepFailed(self.pipeline_name, step, self.last_run_exception, runner.LAST_RUN_METRICS) from r_ex
        finally:
            signals.raise_if_signalled()

    def _run_f_in_pool(self, run_f: Callable[..., Any], config: PoolRunnerConfiguration) -> int:

        def _run(_: Any) -> TRunMetrics:
            rv = run_f()
            if isinstance(rv, TRunMetrics):
                return rv
            if isinstance(rv, int):
                pending = rv
            else:
                pending = 1
            return TRunMetrics(False, False, int(pending))

        # run the fun
        ec = runner.run_pool(config, _run)
        # ec > 0 - signalled
        # -1 - runner was not able to start

        if runner.LAST_RUN_METRICS is not None and runner.LAST_RUN_METRICS.has_failed:
            raise self.last_run_exception
        return ec

    def _get_destination_client_initial_config(self, credentials: Any = None) -> DestinationClientConfiguration:
        if not self.destination:
            raise PipelineConfigMissing(
                self.pipeline_name,
                "destination",
                "load",
                "Please provide `destination` argument to `pipeline`, `run` or `load` method directly or via .dlt config.toml file or environment variable."
            )
        # create initial destination client config
        client_spec = self.destination.spec()
        if issubclass(client_spec, DestinationClientDwhConfiguration):
            # client support schemas and datasets
            return client_spec(dataset_name=self.dataset_name, default_schema_name=self.default_schema_name, credentials=credentials or self.credentials)
        else:
            return client_spec(credentials=credentials or self.credentials)

    def _get_destination_client(self, schema: Schema, initial_config: DestinationClientConfiguration = None) -> JobClientBase:
        try:
            # config is not provided then get it with injected credentials
            if not initial_config:
                initial_config = self._get_destination_client_initial_config()
            return self.destination.client(schema, initial_config)
        except ImportError:
            client_spec = self.destination.spec()
            raise MissingDependencyException(
                f"{client_spec.destination_name} destination",
                [f"{logger.DLT_PKG_NAME}[{client_spec.destination_name}]"],
                "Dependencies for specific destinations are available as extras of python-dlt"
            )

    def _get_destination_capabilities(self) -> DestinationCapabilitiesContext:
        if not self.destination:
                raise PipelineConfigMissing(
                    self.pipeline_name,
                    "destination",
                    "normalize",
                    "Please provide `destination` argument to `pipeline`, `run` or `load` method directly or via .dlt config.toml file or environment variable."
                )
        return self.destination.capabilities()

    def _validate_pipeline_name(self) -> None:
        try:
            FileStorage.validate_file_name_component(self.pipeline_name)
        except ValueError as ve_ex:
            raise InvalidPipelineName(self.pipeline_name, str(ve_ex))

    def _make_schema_with_default_name(self) -> Schema:
        # make a schema from the pipeline name using the name normalizer
        return Schema(self.pipeline_name, normalize_name=True)

    def _validate_dataset_name(self, dataset_name: str) -> None:
        normalized_name = self._default_naming.normalize_schema_name(dataset_name)
        if normalized_name != dataset_name:
            raise InvalidDatasetName(dataset_name, normalized_name)

    def _set_dataset_name(self, dataset_name: str) -> None:
        if not dataset_name:
            if not self.dataset_name:
                # set default dataset name from pipeline name
                dataset_name = self._default_naming.normalize_schema_name(self.pipeline_name)
            else:
                return

        # in case of full refresh add unique suffix
        if self.full_refresh:
            dataset_name += self._pipeline_instance_id
        self._validate_dataset_name(dataset_name)
        self.dataset_name = dataset_name

    def _set_default_schema_name(self, schema: Schema) -> None:
        assert self.default_schema_name is None
        self.default_schema_name = schema.name

    @with_state_sync()
    def _inject_schema(self, schema: Schema) -> None:
        """Injects a schema into the pipeline. Existing schema will be overwritten"""
        self._schema_storage.save_schema(schema)
        if not self.default_schema_name:
            self._set_default_schema_name(schema)

    def _get_load_info(self, load: Load) -> LoadInfo:
        # TODO: Load must provide a clear interface to get last loads and metrics
        # TODO: LoadInfo must hold many packages with different schemas and datasets
        load_ids = load._processed_load_ids
        failed_packages: Dict[str, Sequence[Tuple[str, str]]] = {}
        for load_id, metrics in load_ids.items():
            if metrics:
                # get failed packages only when package finished
                failed_packages[load_id] = self.list_failed_jobs_in_package(load_id)
        dataset_name = None
        if isinstance(load.initial_client_config, DestinationClientDwhConfiguration):
            dataset_name = load.initial_client_config.dataset_name

        return LoadInfo(
            self,
            load.initial_client_config.destination_name,
            str(load.initial_client_config.credentials),
            dataset_name,
            {load_id: bool(metrics) for load_id, metrics in load_ids.items()},
            failed_packages
        )

    def _get_state(self, restore_from_destination: bool = False) -> TPipelineState:
        try:
            state: TPipelineState = json.loads(self._pipeline_storage.load(Pipeline.STATE_FILE))
        except FileNotFoundError:
            if restore_from_destination:
                state = self._restore_state_from_destination()
            else:
                state = None

        # initial state
        if not state:
            state = {
                "_state_version": 0,
                "_state_engine_version": STATE_ENGINE_VERSION
            }
        return state

    def _optional_sql_job_client(self, schema_name: str) -> Optional[SqlJobClientBase]:
        try:
            return self._sql_job_client(Schema(schema_name))
        except PipelineConfigMissing as pip_ex:
            # fallback to regular init if destination not configured
            logger.info(f"Sql Client not available: {pip_ex}")
        except SqlClientNotAvailable:
            # fallback is sql client not available for destination
            logger.info("Client not available because destination does not support sql client")
        except ConfigFieldMissingException:
            # probably credentials are missing
            logger.info("Client not available due to missing credentials")
        return None

    def _restore_state_from_destination(self) -> Optional[TPipelineState]:
        # if state is not present locally, take the state from the destination
        dataset_name = self.dataset_name
        try:
            # patch the default schema name so we load exactly in the dataset
            assert self.default_schema_name is None
            self.default_schema_name = dataset_name
            job_client = self._optional_sql_job_client(dataset_name)
            if job_client:
                with job_client.sql_client as sql_client:
                    state = load_state_from_destination(self.pipeline_name, sql_client)
                    if state is None:
                        logger.info(f"The state was not found in the destination {self.destination.__name__}:{dataset_name}")
                    else:
                        logger.info(f"The state was restored from the destination {self.destination.__name__}:{dataset_name}")
                    return state
            else:
                return None
        finally:
            self.default_schema_name = None

    def _restore_schemas_from_destination(self) -> None:
        # check which schemas are present in the pipeline and restore missing schemas
        # existing_schemas = set(self._schema_storage.list_schemas())
        for schema_name in self.schema_names:
            if not self._schema_storage.has_schema(schema_name):
                # assume client always exist
                with self._optional_sql_job_client(schema_name) as job_client:
                    schema_info = job_client.get_newest_schema_from_storage()
                    if schema_info is None:
                        logger.info(f"The schema {schema_name} was not found in the destination {self.destination.__name__}:{job_client.sql_client.default_dataset_name}. Default schema was placed instead")
                        # try to import schema
                        try:
                            self._schema_storage.load_schema(schema_name)
                        except FileNotFoundError:
                            self._schema_storage.save_schema(job_client.schema)
                    else:
                        logger.info(f"The schema {schema_name} was restored from the destination {self.destination.__name__}:{job_client.sql_client.default_dataset_name}")
                        schema = Schema.from_dict(json.loads(schema_info.schema))
                        self._schema_storage.save_schema(schema)

    @contextmanager
    def _managed_state(self, *, restore_from_destination: bool = False, extract_state: bool = False) -> Iterator[TPipelineState]:
        # load current state
        state = self._get_state(restore_from_destination)

        # write props to pipeline variables
        for prop in Pipeline.STATE_PROPS:
            if prop in state and not prop.startswith("_"):
                setattr(self, prop, state.get(prop))
        if "destination" in state:
            self.destination = DestinationReference.from_name(self.destination)

        try:
            yield state
        except Exception:
            # currently do nothing - state is not preserved in memory, only saved
            raise
        else:
            # update state props
            for prop in Pipeline.STATE_PROPS:
                if not prop.startswith("_"):
                    state[prop] = getattr(self, prop)  # type: ignore
            if self.destination:
                state["destination"] = self.destination.__name__
            state["schema_names"] = self._schema_storage.list_schemas()

            backup_state = self._get_state()
            merged_state = merge_state_if_changed(backup_state, state)
            # state was changed or current state was not yet extracted and there's request to extract state
            if (merged_state or "_last_extracted_at" not in state) and extract_state:
                merged_state = merged_state or state
                merged_state["_last_extracted_at"] = pendulum.now()
                # this will extract the state into current load package and update the schema with the _dlt_pipeline_state table
                # note: the schema will be persisted because the schema saving decorator is over the state manager decorator for extract
                state_source = DltSource("pipeline_state", self.default_schema, [state_resource(merged_state)])
                self._iterate_source(state_source, self.default_schema, 1, 1)

            if not extract_state and merged_state:
                merged_state.pop("_last_extracted_at", None)

            # if state was changed, save it
            if merged_state:
                self._pipeline_storage.save(Pipeline.STATE_FILE, json.dumps(merged_state))



from .specs.base_configuration import configspec, is_valid_hint  # noqa: F401
from .resolve import resolve_configuration, inject_namespace  # noqa: F401
from .inject import with_config, last_config, get_fun_spec

from .exceptions import (  # noqa: F401
    ConfigFieldMissingException, ConfigValueCannotBeCoercedException, ConfigIntegrityException, ConfigFileNotFoundException)



from typing import Type

from dlt.common.schema.schema import Schema
from dlt.common.configuration import with_config
from dlt.common.configuration.accessors import config
from dlt.common.data_writers.escape import escape_redshift_identifier, escape_redshift_literal
from dlt.common.destination import DestinationCapabilitiesContext, JobClientBase, DestinationClientConfiguration

from dlt.destinations.redshift.configuration import RedshiftClientConfiguration


@with_config(spec=RedshiftClientConfiguration, namespaces=("destination", "redshift",))
def _configure(config: RedshiftClientConfiguration = config.value) -> RedshiftClientConfiguration:
    return config


def capabilities() -> DestinationCapabilitiesContext:
    caps = DestinationCapabilitiesContext()
    caps.update({
        "preferred_loader_file_format": "insert_values",
        "supported_loader_file_formats": ["insert_values"],
        "escape_identifier": escape_redshift_identifier,
        "escape_literal": escape_redshift_literal,
        "max_identifier_length": 127,
        "max_column_identifier_length": 127,
        "max_query_length": 16 * 1024 * 1024,
        "is_max_query_length_in_bytes": True,
        "max_text_data_type_length": 65535,
        "is_max_text_data_type_length_in_bytes": True
    })
    return caps


def client(schema: Schema, initial_config: DestinationClientConfiguration = config.value) -> JobClientBase:
    # import client when creating instance so capabilities and config specs can be accessed without dependencies installed
    from dlt.destinations.redshift.redshift import RedshiftClient

    return RedshiftClient(schema, _configure(initial_config))  # type: ignore


def spec() -> Type[DestinationClientConfiguration]:
    return RedshiftClientConfiguration



import os
import tomlkit
from tomlkit.items import Item as TOMLItem
from tomlkit.container import Container as TOMLContainer
from typing import Any, Optional, Tuple, Type, Union

from dlt.common.typing import StrAny

from .provider import ConfigProvider, ConfigProviderException


class TomlProvider(ConfigProvider):

    def __init__(self, file_name: str, project_dir: str = None) -> None:
        self._file_name = file_name
        self._toml_path = os.path.join(project_dir or os.path.abspath(os.path.join(".", ".dlt")), file_name)
        try:
            self._toml = self._read_toml(self._toml_path)
        except Exception as ex:
            raise TomlProviderReadException(self.name, file_name, self._toml_path, str(ex))

    @staticmethod
    def get_key_name(key: str, *namespaces: str) -> str:
        # env key is always upper case
        if namespaces:
            namespaces = filter(lambda x: bool(x), namespaces)  # type: ignore
            env_key = ".".join((*namespaces, key))
        else:
            env_key = key
        return env_key

    def get_value(self, key: str, hint: Type[Any], *namespaces: str) -> Tuple[Optional[Any], str]:
        full_path = namespaces + (key,)
        full_key = self.get_key_name(key, *namespaces)
        node: Union[TOMLContainer, TOMLItem] = self._toml
        try:
            for k in  full_path:
                if not isinstance(node, dict):
                    raise KeyError(k)
                node = node[k]
            return node, full_key
        except KeyError:
            return None, full_key

    @property
    def supports_namespaces(self) -> bool:
        return True

    def _write_toml(self) -> None:
        with open(self._toml_path, "w", encoding="utf-8") as f:
            tomlkit.dump(self._toml, f)

    @staticmethod
    def _read_toml(toml_path: str) -> tomlkit.TOMLDocument:
        if os.path.isfile(toml_path):
            with open(toml_path, "r", encoding="utf-8") as f:
                # use whitespace preserving parser
                return tomlkit.load(f)
        else:
            return tomlkit.document()


class ConfigTomlProvider(TomlProvider):

    def __init__(self, project_dir: str = None) -> None:
        super().__init__("config.toml", project_dir)

    @property
    def name(self) -> str:
        return "Pipeline config.toml"

    @property
    def supports_secrets(self) -> bool:
        return False



class SecretsTomlProvider(TomlProvider):

    def __init__(self, project_dir: str = None) -> None:
        super().__init__("secrets.toml", project_dir)

    @property
    def name(self) -> str:
        return "Pipeline secrets.toml"

    @property
    def supports_secrets(self) -> bool:
        return True


class TomlProviderReadException(ConfigProviderException):
    def __init__(self, provider_name: str, file_name: str, full_path: str, toml_exception: str) -> None:
        self.file_name = file_name
        self.full_path = full_path
        msg = f"A problem encountered when loading {provider_name} from {full_path}:\n"
        msg += toml_exception
        super().__init__(provider_name, msg)



from typing import Any, Iterator

import dlt
from dlt.common.typing import StrAny, TDataItem, TDataItems
from dlt.common.time import timestamp_within
from dlt.extract.source import DltResource


@dlt.source
def rasa(
    data_from: DltResource,
    /,
    source_env: str = None,
    initial_timestamp: float = None,
    end_timestamp: float = None,
    store_last_timestamp: bool = True
) -> Any:
    """Transforms the base resource provided in `data_from` into a rasa tracker store raw dataset where each event type get it's own table.
    The resource is a stream resource and it generates tables dynamically from data. The source uses `rasa.schema.yaml` file to initialize the schema

    Args:
        data_from (DltResource): the base resource with raw events data
        source_env (str, optional): the "environment" field value. Defaults to None.
        initial_timestamp (float, optional): filters in only the events past the timestamp. Defaults to None.
        end_timestamp (float, optional): filters in only the events before the timestamp. Defaults to None.
        store_last_timestamp (bool, optional): tells the source if it should preserve last processed timestamp value and resume when it is run again. Defaults to True.

    Returns:
        DltSource: a source with single resource called "events"
    """

    # the data_from can be a file, database table (see sql_query.py), kafka topic, rabbitmq queue etc.
    @dlt.transformer(data_from=data_from)
    def events(source_events: TDataItems) -> Iterator[TDataItem]:
        # recover start_timestamp from state if given
        if store_last_timestamp:
            start_timestamp = max(initial_timestamp or 0, dlt.state().get("start_timestamp", 0))
        # we expect tracker store events here
        last_timestamp: int = None

        def _proc_event(source_event: TDataItem) -> Iterator[TDataItem]:
            nonlocal last_timestamp

            # must be a dict
            assert isinstance(source_event, dict)
            # filter out events
            if timestamp_within(source_event["timestamp"], start_timestamp, end_timestamp):
                # yield tracker table with all-event index
                event_type = source_event["event"]
                last_timestamp = source_event["timestamp"]
                event = {
                    "sender_id": source_event["sender_id"],
                    "timestamp": last_timestamp,
                    "event": event_type
                }
                if source_env:
                    event["source"] = source_env
                if "model_id" in source_event:
                    event["model_id"] = source_event["model_id"]

                yield dlt.with_table_name(event, "event")
                # yield original event
                yield dlt.with_table_name(source_event, "event_" + event_type)

        # events may be a single item or list of items
        if isinstance(source_events, list):
            for source_event in source_events:
                yield from _proc_event(source_event)
        else:
            yield from _proc_event(source_events)

        # write state so we can restart
        if store_last_timestamp and last_timestamp:
            dlt.state()["start_timestamp"] = last_timestamp

    return events



"""dlt

How to create a data loading pipeline with dlt in 3 seconds:

    1. Write a pipeline script
    >>> import dlt
    >>> dlt.run(source=my_complicated_json, destination="duckdb")

    2. Run your pipeline script
    $ python my_pipeline.py

    3. See and use your data
    $ dlt pipeline show my_pipeline.py

    This will auto-generate and run a Streamlit app where you can see the data and the schema


Or start with our pipeline template with sample chess.com data to bigquery

    $ dlt init chess bigquery

For more detailed info, see https://dlthub.com/docs
"""

from dlt.pipeline import pipeline as _pipeline, run, attach, Pipeline
from dlt.pipeline.state import state
from dlt.extract.decorators import source, resource, transformer, defer
from dlt.extract.source import with_table_name
from dlt.common.schema import Schema
from dlt.common.configuration.accessors import config, secrets
from dlt.common.typing import TSecretValue as _TSecretValue
from dlt.common.configuration.specs import CredentialsConfiguration as _CredentialsConfiguration

pipeline = _pipeline

TSecretValue = _TSecretValue
"When typing source/resource function arguments indicates that given argument is a secret and should be taken from dlt.secrets. The value itself is a string"

TCredentials = _CredentialsConfiguration
"When typing source/resource function arguments indicates that given argument represents credentials and should be taken from dlt.secrets. Credentials may be string, dictionaries or any other types."

from dlt.__version__ import __version__





from .run_configuration import RunConfiguration  # noqa: F401
from .base_configuration import BaseConfiguration, CredentialsConfiguration, ContainerInjectableContext  # noqa: F401
from .normalize_volume_configuration import NormalizeVolumeConfiguration  # noqa: F401
from .load_volume_configuration import LoadVolumeConfiguration  # noqa: F401
from .schema_volume_configuration import SchemaVolumeConfiguration, TSchemaFileFormat  # noqa: F401
from .pool_runner_configuration import PoolRunnerConfiguration, TPoolType  # noqa: F401
from .gcp_client_credentials import GcpClientCredentials, GcpClientCredentialsWithDefault  # noqa: F401
from .postgres_credentials import PostgresCredentials, ConnectionStringCredentials  # noqa: F401
from .config_namespace_context import ConfigNamespacesContext  # noqa: F401


import dataclasses
from os import environ
from typing import List, Optional, Type

from dlt.common.typing import StrAny, TSecretValue
from dlt.common.configuration import resolve_configuration, configspec
from dlt.common.configuration.providers import EnvironProvider
from dlt.common.configuration.specs import RunConfiguration, PoolRunnerConfiguration, TPoolType, PostgresCredentials, GcpClientCredentials


@configspec
class DBTRunnerConfiguration(RunConfiguration, PoolRunnerConfiguration):
    pool_type: TPoolType = "none"
    stop_after_runs: int = 1
    package_volume_path: str = "/var/local/app"
    package_repository_url: str = "https://github.com/scale-vector/rasa_semantic_schema_customization.git"
    package_repository_branch: Optional[str] = None
    package_repository_ssh_key: TSecretValue = TSecretValue("")  # the default is empty value which will disable custom SSH KEY
    package_profiles_dir: str = "."
    package_profile_prefix: str = "rasa_semantic_schema"
    package_source_tests_selector: str = "tag:prerequisites"
    package_additional_vars: Optional[StrAny] = None
    package_run_params: List[str] = dataclasses.field(default_factory=lambda: ["--fail-fast"])
    auto_full_refresh_when_out_of_sync: bool = True

    source_schema_prefix: str = None
    dest_schema_prefix: Optional[str] = None

    def on_resolved(self) -> None:
        if self.package_repository_ssh_key and self.package_repository_ssh_key[-1] != "\n":
            # must end with new line, otherwise won't be parsed by Crypto
            self.package_repository_ssh_key = TSecretValue(self.package_repository_ssh_key + "\n")
        if self.stop_after_runs != 1:
            # always stop after one run
            self.stop_after_runs = 1


def gen_configuration_variant(explicit_values: StrAny = None) -> DBTRunnerConfiguration:
    # derive concrete config depending on env vars present
    DBTRunnerConfigurationImpl: Type[DBTRunnerConfiguration]
    environ = EnvironProvider()

    source_schema_prefix: str = environ.get_value("dataset_name", type(str))  # type: ignore

    if environ.get_value("project_id", type(str), GcpClientCredentials.__namespace__):
        @configspec
        class DBTRunnerConfigurationPostgres(PostgresCredentials, DBTRunnerConfiguration):
            SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerConfigurationImpl = DBTRunnerConfigurationPostgres

    else:
        @configspec
        class DBTRunnerConfigurationGcp(GcpClientCredentials, DBTRunnerConfiguration):
            SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerConfigurationImpl = DBTRunnerConfigurationGcp

    return resolve_configuration(DBTRunnerConfigurationImpl(), explicit_value=explicit_values)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dlt',
 'dlt.cli',
 'dlt.common',
 'dlt.common.configuration',
 'dlt.common.configuration.providers',
 'dlt.common.configuration.specs',
 'dlt.common.data_writers',
 'dlt.common.normalizers.json',
 'dlt.common.normalizers.names',
 'dlt.common.reflection',
 'dlt.common.runners',
 'dlt.common.schema',
 'dlt.common.storages',
 'dlt.dbt_runner',
 'dlt.destinations',
 'dlt.destinations.bigquery',
 'dlt.destinations.dummy',
 'dlt.destinations.postgres',
 'dlt.destinations.redshift',
 'dlt.extract',
 'dlt.helpers',
 'dlt.load',
 'dlt.normalize',
 'dlt.pipeline',
 'dlt.reflection',
 'examples',
 'examples.chess',
 'examples.schemas',
 'examples.sources',
 'examples.sources.rasa']

package_data = \
{'': ['*'],
 'examples': ['.dlt/example.secrets.toml',
              'data/*',
              'data/rasa_trackers/*',
              'data/singer_taps/*'],
 'examples.chess': ['.dlt/*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.3.5,<2.0.0',
 'astunparse>=1.6.3,<2.0.0',
 'asyncstdlib>=3.10.5,<4.0.0',
 'cachetools>=5.2.0,<6.0.0',
 'click>=8.1.3,<9.0.0',
 'hexbytes>=0.2.2,<0.3.0',
 'humanize>=4.4.0,<5.0.0',
 'json-logging==1.4.1rc0',
 'jsonlines>=2.0.0,<3.0.0',
 'makefun>=1.15.0,<2.0.0',
 'pathvalidate>=2.5.2,<3.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'prometheus-client>=0.11.0,<0.12.0',
 'requests>=2.26.0,<3.0.0',
 'requirements-parser>=0.5.0,<0.6.0',
 'semver>=2.13.0,<3.0.0',
 'sentry-sdk>=1.4.3,<2.0.0',
 'setuptools>=65.6.0,<66.0.0',
 'simplejson>=3.17.5,<4.0.0',
 'tomlkit>=0.11.3,<0.12.0',
 'typing-extensions>=4.0.0,<5.0.0',
 'tzdata>=2022.1,<2023.0']

extras_require = \
{':extra == "dbt"': ['GitPython>=3.1.26,<4.0.0'],
 'bigquery': ['pyarrow>=8.0.0,<9.0.0'],
 'bigquery:python_version < "3.11"': ['grpcio>=1.50.0,<2.0.0',
                                      'google-cloud-bigquery>=2.26.0,<3.0.0',
                                      'google-cloud-bigquery-storage>=2.13.0,<3.0.0'],
 'dbt': ['dbt-core>=1.1.0,<1.2.0', 'dbt-redshift>=1.0.0,<1.2.0'],
 'dbt:python_version < "3.11"': ['dbt-bigquery>=1.0.0,<1.2.0'],
 'gcp': ['pyarrow>=8.0.0,<9.0.0'],
 'gcp:python_version < "3.11"': ['grpcio>=1.50.0,<2.0.0',
                                 'google-cloud-bigquery>=2.26.0,<3.0.0',
                                 'google-cloud-bigquery-storage>=2.13.0,<3.0.0'],
 'postgres': ['psycopg2-binary>=2.9.1,<3.0.0'],
 'postgres:platform_python_implementation == "PyPy"': ['psycopg2cffi>=2.9.0,<3.0.0'],
 'redshift': ['psycopg2-binary>=2.9.1,<3.0.0'],
 'redshift:platform_python_implementation == "PyPy"': ['psycopg2cffi>=2.9.0,<3.0.0']}

entry_points = \
{'console_scripts': ['dlt = dlt.cli._dlt:main']}

setup_kwargs = {
    'name': 'python-dlt',
    'version': '0.2.0a2',
    'description': 'DLT is an open-source python-native scalable data loading framework that does not require any devops efforts to run.',
    'long_description': '# Quickstart Guide: Data Load Tool (DLT)\n\n## **TL;DR: This guide shows you how to load a JSON document into Google BigQuery using DLT.**\n\n![](docs/DLT-Pacman-Big.gif)\n\n*Please open a pull request [here](https://github.com/scale-vector/dlt/edit/master/QUICKSTART.md) if there is something you can improve about this quickstart.*\n\n## Grab the demo\n\nClone the example repository:\n```\ngit clone https://github.com/scale-vector/dlt-quickstart-example.git\n```\n\nEnter the directory:\n```\ncd dlt-quickstart-example\n```\n\nOpen the files in your favorite IDE / text editor:\n- `data.json` (i.e. the JSON document you will load)\n- `credentials.json` (i.e. contains the credentials to our demo Google BigQuery warehouse)\n- `quickstart.py` (i.e. the script that uses DLT)\n\n## Set up a virtual environment\n\nEnsure you are using either Python 3.8 or 3.9:\n```\npython3 --version\n```\n\nCreate a new virtual environment:\n```\npython3 -m venv ./env\n```\n\nActivate the virtual environment:\n```\nsource ./env/bin/activate\n```\n\n## Install DLT and support for the target data warehouse\n\nInstall DLT using pip:\n```\npip3 install -U python-dlt\n```\n\nInstall support for Google BigQuery:\n```\npip3 install -U python-dlt[gcp]\n```\n\n## Understanding the code\n\n1. Configure DLT\n\n2. Create a DLT pipeline\n\n3. Load the data from the JSON document\n\n4. Pass the data to the DLT pipeline\n\n5. Use DLT to load the data\n\n## Running the code\n\nRun the quickstart script in `/examples` folder:\n\n```\npython3 quickstart.py\n```\n\nInspect `schema.yml` that has been printed by the script or the generated file:\n```\nvim schema.yml\n```\n\nSee results of querying the Google BigQuery table:\n\n`json_doc` table\n\n```\nSELECT * FROM `{schema_prefix}_example.json_doc`\n```\n```\n{  "name": "Ana",  "age": "30",  "id": "456",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "5b018c1ba3364279a0ca1a231fbd8d90"}\n{  "name": "Bob",  "age": "30",  "id": "455",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "afc8506472a14a529bf3e6ebba3e0a9e"}\n```\n\n`json_doc__children` table\n\n```\nSELECT * FROM `{schema_prefix}_example.json_doc__children` LIMIT 1000\n```\n```\n    # {"name": "Bill", "id": "625", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "0", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",\n    #   "_dlt_id": "7993452627a98814cc7091f2c51faf5c"}\n    # {"name": "Bill", "id": "625", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "0", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",\n    #   "_dlt_id": "9a2fd144227e70e3aa09467e2358f934"}\n    # {"name": "Dave", "id": "621", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "1", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",\n    #   "_dlt_id": "28002ed6792470ea8caf2d6b6393b4f9"}\n    # {"name": "Elli", "id": "591", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "1", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",\n    #   "_dlt_id": "d18172353fba1a492c739a7789a786cf"}\n```\n\nJoining the two tables above on autogenerated keys (i.e. `p._record_hash = c._parent_hash`)\n\n```\nselect p.name, p.age, p.id as parent_id,\n            c.name as child_name, c.id as child_id, c._dlt_list_idx as child_order_in_list\n        from `{schema_prefix}_example.json_doc` as p\n        left join `{schema_prefix}_example.json_doc__children`  as c\n            on p._dlt_id = c._dlt_parent_id\n```\n```\n    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}\n    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Elli",  "child_id": "591",  "child_order_in_list": "1"}\n    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}\n    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Dave",  "child_id": "621",  "child_order_in_list": "1"}\n```\n\n## Next steps\n\n1. Replace `data.json` with data you want to explore\n\n2. Check that the inferred types are correct in `schema.yml`\n\n3. Set up your own Google BigQuery warehouse (and replace the credentials)\n\n4. Use this new clean staging layer as the starting point for a semantic layer / analytical model (e.g. using dbt)',
    'author': 'ScaleVector',
    'author_email': 'services@scalevector.ai',
    'maintainer': 'Marcin Rudolf',
    'maintainer_email': 'marcin@scalevector.ai',
    'url': 'https://github.com/scale-vector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Iterator

import dlt

from dlt.common import json
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.runners.venv import Venv
from dlt.common.runners.stdout import iter_stdout
from dlt.common.typing import DictStrAny


# create a standalone resource from a pipe to another process stdout. deselect the source by default as it will be used mostly as a data source to a transformer
@dlt.resource(selected=False, spec=BaseConfiguration)
def json_stdout(venv: Venv, command: str, *script_args: Any) -> Iterator[DictStrAny]:
    """Create a standalone resource from a pipe to another process stdout. Yields line by line. Parses lines as json."""
    # use pipe iterator and mapping function to get dict iterator from pipe
    yield from map(lambda s: json.loads(s), iter_stdout(venv, command, *script_args))  # type: ignore



from typing import Sequence, Union, cast
from dlt.common.schema import Schema
from dlt.common.schema.typing import TColumnSchema, TWriteDisposition

from dlt.common.typing import TSecretValue, Any
from dlt.common.configuration import with_config
from dlt.common.configuration.container import Container
from dlt.common.configuration.inject import get_orig_args
from dlt.common.destination import DestinationReference, TDestinationReferenceArg
from dlt.common.pipeline import LoadInfo, PipelineContext, get_default_working_dir

from dlt.pipeline.configuration import PipelineConfiguration
from dlt.pipeline.pipeline import Pipeline


@with_config(spec=PipelineConfiguration, auto_namespace=True)
def pipeline(
    pipeline_name: str = None,
    working_dir: str = None,
    pipeline_salt: TSecretValue = None,
    destination: TDestinationReferenceArg = None,
    dataset_name: str = None,
    import_schema_path: str = None,
    export_schema_path: str = None,
    full_refresh: bool = False,
    restore_from_destination: bool = False,
    credentials: Any = None,
    **kwargs: Any
) -> Pipeline:
    # call without arguments returns current pipeline
    orig_args = get_orig_args(**kwargs)  # original (*args, **kwargs)
    # is any of the arguments different from defaults
    has_arguments = bool(orig_args[0]) or any(orig_args[1].values())

    if not has_arguments:
        context = Container()[PipelineContext]
        # if pipeline instance is already active then return it, otherwise create a new one
        if context.is_active():
            return cast(Pipeline, context.pipeline())
        else:
            pass

    # if working_dir not provided use temp folder
    if not working_dir:
        working_dir = get_default_working_dir()

    destination = DestinationReference.from_name(destination or kwargs["destination_name"])
    # create new pipeline instance
    p = Pipeline(
        pipeline_name,
        working_dir,
        pipeline_salt,
        destination,
        dataset_name,
        credentials,
        import_schema_path,
        export_schema_path,
        full_refresh,
        False,
        restore_from_destination,
        kwargs["runtime"])
    # set it as current pipeline
    Container()[PipelineContext].activate(p)

    return p


@with_config(spec=PipelineConfiguration, auto_namespace=True)
def attach(
    pipeline_name: str = None,
    working_dir: str = None,
    pipeline_salt: TSecretValue = None,
    full_refresh: bool = False,
    **kwargs: Any
) -> Pipeline:
    # if working_dir not provided use temp folder
    if not working_dir:
        working_dir = get_default_working_dir()
    # create new pipeline instance
    p = Pipeline(pipeline_name, working_dir, pipeline_salt, None, None, None, None, None, full_refresh, True, False, kwargs["runtime"])
    # set it as current pipeline
    Container()[PipelineContext].activate(p)
    return p


# setup default pipeline in the container
Container()[PipelineContext] = PipelineContext(pipeline)


def run(
    data: Any,
    *,
    destination: TDestinationReferenceArg = None,
    dataset_name: str = None,
    credentials: Any = None,
    table_name: str = None,
    write_disposition: TWriteDisposition = None,
    columns: Sequence[TColumnSchema] = None,
    schema: Schema = None
) -> LoadInfo:
    destination = DestinationReference.from_name(destination)
    return pipeline().run(
        data,
        destination=destination,
        dataset_name=dataset_name,
        credentials=credentials,
        table_name=table_name,
        write_disposition=write_disposition,
        columns=columns,
        schema=schema
    )




import jsonlines
import itertools
from typing import Iterator, List, Sequence

import dlt
from dlt.common import json
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.typing import StrAny, StrOrBytesPath


def chunk_jsonl(path: StrOrBytesPath, chunk_size: int = 20) -> Iterator[List[StrAny]]:
    print(path)
    with open(path, "r", encoding="utf-8") as f:
        _iter = jsonlines.Reader(f, loads=json.loads)
        if chunk_size == 1:
            yield from _iter
        else:
            while True:
                chunk = list(itertools.islice(_iter, chunk_size))
                if chunk:
                    yield chunk
                else:
                    break

# TODO: explicit values must take precedence
jsonl_file = dlt.resource(chunk_jsonl, name="jsonl", spec=BaseConfiguration)

@dlt.resource(name="jsonl")
def jsonl_files(paths: Sequence[StrOrBytesPath], chunk_size: int = 20) -> Iterator[List[StrAny]]:
    for path in paths:
        yield from chunk_jsonl(path, chunk_size)



# from apiclient.discovery import build
# from oauth2client.service_account import ServiceAccountCredentials
# from googleapiclient.http import MediaIoBaseDownload
# import io
# from typing import Any, Iterator
# import csv

# from dlt.common.typing import DictStrAny, StrAny
# from dlt.common.schema import Schema

# from dlt.pipeline import Pipeline, GCPPipelineCredentials

# SCOPES = ['https://www.googleapis.com/auth/drive']
# # KEY_FILE_LOCATION = '/Users/adrian/PycharmProjects/sv/dlt/temp/scalevector-1235ac340b0b.json'
# KEY_FILE_LOCATION = '_secrets/scalevector-1235ac340b0b.json'


# def _initialize_drive() -> Any:
#     """Initializes an drive service object.

#     Returns:
#     An authorized drive service object.
#     """
#     credentials = ServiceAccountCredentials.from_json_keyfile_name(
#         KEY_FILE_LOCATION, SCOPES)

#     # Build the service object.
#     service = build('drive', 'v3', credentials=credentials)

#     return service


# # TODO: consider using https://pythonhosted.org/PyDrive/index.html or https://github.com/wkentaro/gdown
# def download_csv_as_json(file_id: str, csv_options: StrAny = None) -> Iterator[DictStrAny]:
#     if csv_options is None:
#         csv_options = {}

#     drive_service = _initialize_drive()
#     request = drive_service.files().get_media(fileId=file_id)
#     fh = io.BytesIO()
#     downloader = MediaIoBaseDownload(fh, request)
#     done = False
#     while done is False:
#         status, done = downloader.next_chunk()
#         print("Download %d%%." % int(status.progress() * 100))
#     rows = fh.getvalue().decode("utf-8")
#     return csv.DictReader(io.StringIO(rows), **csv_options)


# if __name__ == "__main__":
#     file_id = '1dkJYb-WVVseZIQLAY0jh_o0s7Bas4Gp5'
#     options = {'delimiter': ';'}
#     iter_d = download_csv_as_json(file_id, csv_options=options)

#     # LOADING DETAILS
#     table_n = 'people'
#     schema_prefix = 'drive'
#     schema_source_suffix = 'csv'

#     # you authenticate by passing a credential, such as RDBMS host/user/port/pass or gcp service account json.
#     gcp_credential_json_file_path = KEY_FILE_LOCATION

#     # SCHEMA CREATION
#     data_schema = None
#     data_schema_file_path = f"examples/schemas/inferred_drive_csv_{file_id}.schema.yml"

#     credentials = GCPPipelineCredentials.from_services_file(gcp_credential_json_file_path, schema_prefix)

#     pipeline = Pipeline(schema_source_suffix)

#     schema: Schema = None
#     # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
#     # TODO: schema below needs is not yet finished, more exploration needed
#     # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
#     pipeline.create_pipeline(credentials, schema=schema)
#     # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
#     # Pipeline::restore_pipeline
#     print(pipeline.root_path)
#     # /private/var/folders/4q/gzmrslrs03j587990jcqt23h0000gn/T/tmp5e17osrg/

#     # and extract it
#     pipeline.extract(iter_d, table_name=table_n)
#     pipeline.flush()

#     # save schema
#     schema = pipeline.get_default_schema()
#     schema_yaml = schema.to_pretty_yaml()
#     f = open(data_schema_file_path, "a", encoding="utf-8")
#     f.write(schema_yaml)
#     f.close()



# from dlt.pipeline import Pipeline, GCPPipelineCredentials

# if __name__ == '__main__':
#     credentials = GCPPipelineCredentials.default_credentials("mainnet_3")
#     # credentials = GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "mainnet_3")
#     # credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_2", "loader", "3.73.90.3")

#     pipeline = Pipeline("example")
#     # restore pipeline from the working directory
#     # working directory contains pipeline full state after it was created
#     # it is possible to restore pipeline in many concurrently running components ie. several extractors, separate normalize and load
#     # pipeline.restore_pipeline(credentials, "/tmp/tmp724aveoc/")
#     pipeline.restore_pipeline(credentials, "/tmp/tmpqgouqwoo")

#     # the code below will normalize and load anything that was extracted by extractor process(es)

#     # do we have anything to normalize
#     print(pipeline.list_extracted_loads())

#     # just finalize
#     pipeline.flush()


import dlt

from examples.sources.google_sheets import google_spreadsheet

dlt.pipeline(destination="bigquery", full_refresh=False)
# see example.secrets.toml to where to put credentials

info = google_spreadsheet("11G95oVZjieRhyGqtQMQqlqpxyvWkRXowKE8CtdLtFaU", ["2022-05", "model_metadata"]).run()
print(info)
